
import thread
import sys
import random

sys.path.append("..")

import code_coverage
import crash_binning
#import pida

import pydbg.pydbg
from pydbg.defines import *

import cuchi
    
####################################################################################################################
class G_Phazz_executor:

    ignore_first_chance = True
    print_logs = True

    ####################################################################################################################
    def log(self, s):
        if self.print_logs:
            print s
        
    ####################################################################################################################
    def handler_access_violation (self, dbg):

        # if the user wants to ignore first chance exceptions then do so.
        if self.ignore_first_chance and dbg.dbg.u.Exception.dwFirstChance:
            return DBG_EXCEPTION_NOT_HANDLED

        crash_bin = crash_binning.crash_binning()
        crash_bin.record_crash(dbg)

        self.log(crash_bin.crash_synopsis())
        dbg.terminate_process()

        # detach
        # return DBG_CONTINUE
        # return DBG_EXCEPTION_HANDLED

    ####################################################################################################################
    def handler_breakpoint (self, dbg):

        if dbg.get_attr("first_breakpoint"):
            return DBG_CONTINUE

        if self.print_bps:
            self.log("debugger hit %08x cc #%d" % (dbg.exception_address, self.cc.num))

        is_function = 0
        if self.module.nodes.has_key(dbg.context.Eip):
            is_function = 1

        self.cc.add(dbg, is_function)

        return DBG_CONTINUE


    ####################################################################################################################
    def set_bps (self):

        basic_blocks = []

        for f in self.module.nodes.values():
            for bb in f.nodes.values():
                basic_blocks.append(bb.ea_start)

        if self.print_setbps:
            self.log("Setting %d breakpoints on basic blocks in main module" % len(basic_blocks))

        self.dbg.bp_set(basic_blocks, restore=self.restore)


    ####################################################################################################################
    def execute (self, args):

        self.restore = False
        self.print_bps = False
        self.print_setbps = False

        self.cc = code_coverage.code_coverage()

        self.dbg = pydbg.pydbg()
        self.dbg.set_callback(EXCEPTION_BREAKPOINT,       self.handler_breakpoint)
        self.dbg.set_callback(EXCEPTION_ACCESS_VIOLATION, self.handler_access_violation)

        #self.dbg.set_callback(LOAD_DLL_DEBUG_EVENT,       self.handler_load_dll)
        #self.dbg.set_callback(USER_CALLBACK_DEBUG_EVENT,  self.handler_user_callback)

        # set the main module name for the code coverage class.
        #self.cc.main_module = self.main

        self.dbg.load(self.filename, args, hidden = True)

        self.set_bps()

        self.dbg.run()

        # self.dbg.terminate_process()
        
    ####################################################################################################################
    def load(self, filename):

        self.filename = filename
        
        cuchi.debug = 0    
        pa = cuchi.module()
        self.module = pa
        
        pa.load(self.filename)
        pa.analyze()

        for f in pa.nodes:
            self.log( "Function at: 0x%x" % f )
            for b in pa.nodes[f].nodes:
                self.log( "  %s" % pa.nodes[f].nodes[b] )

####################################################################################################################

class Organism:

    def __init__(self, gm):
        self.gm = gm
        self.index = -1
        self.fitness = -1
        self.content = ""

    def get_filename(self):
        return self.gm.directory + 'gph%s.txt' % (self.index)
    
    def load_content(self):
        f = open(self.get_filename(), "rb")
        self.content = f.read()
        f.close()
        
    def write_content(self):
        f = open(self.get_filename(), "wb")        
        f.write(self.content)
        f.close()
        
    def get_content(self):
        if(self.content == ""):
            self.load_content()
        return self.content
            
    def calculate_fitness(self):

        if self.gm.read_from_file:
            self.gm.gphx.execute(self.get_filename())
        else:
            self.gm.gphx.execute(self.content)
            
        self.fitness = len(self.gm.gphx.cc.hits)
        return self.fitness
        
    def get_fitness(self):
        if(self.fitness < 0):
            self.calculate_fitness()
        return self.fitness

    def __cmp__(self, other):
        if( self.fitness > other.fitness ):
            return -1
        if( self.fitness < other.fitness ):
            return 1
        if( self.index < other.index ):
            return -1
        if( self.index > other.index ):
            return 1
        return 0
        
####################################################################################################################
    
class Population:
    
    def __init__(self, gm):
        self.gm = gm
        self.organisms = []
        self.total_fitness = 0
        self.select_probability = 0.40
        self.roulette_wheel = False
        self.tournament = True
        
    def append(self, new_org):
        self.organisms.append(new_org)
        self.total_fitness = self.total_fitness + new_org.fitness

    def select_roulette_wheel(self):
        rand_point = random.randint(1, self.total_fitness)
        sum = 0
        for x in xrange(len(self.organisms)):
            sum = sum + self.organisms[x].fitness
            if( sum >= rand_point ):
                return self.organisms[x]
            
        print "no tiene que llegar aca!!"    
        

    def select_tournament(self):
        k = 0
        while k < len(self.organisms):
            if( random.random() < self.select_probability ):
                return self.organisms[k]
            k = k+1
        return self.organisms[ len(self.organisms) - 1 ]

    def select(self):
        if self.roulette_wheel:
            return self.select_roulette_wheel()
        if self.tournament:
            return self.select_tournament()
        
    def sort(self):
        self.organisms.sort()

####################################################################################################################
    
class G_Phazz_modifier:

    def __init__(self):
        self.counter = 0
        self.pop = Population(self)
        self.byteMin = 0x30
        self.byteMax = 0x39
        self.byteList = [' ', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9']
        self.popCount = 20
        self.mutation_probability = 0.70
        self.read_from_file = False
        
    def load_executable(self, filename):
        self.gphx = G_Phazz_executor()
        self.gphx.load(filename)

    def load_initial_content(self, directory, filename):
        self.directory = directory
        f = open(directory + filename, "rb")
        self.initial = Organism(self)
        self.initial.content = f.read()
        f.close()

    def commit(self, org):
        self.counter = self.counter + 1
        org.index = self.counter
        org.write_content()
        org.calculate_fitness()
        
    def mutate(self, org):
        place = random.randint( 0, len(org.get_content()) - 1 )
        value = random.randint( 0, len(self.byteList) - 1 )
        new_org = Organism(self)
        new_org.content = org.content[:place] + self.byteList[value] + org.content[place+1:]
        return new_org
    
    def cross_over(self, org1, org2):
        cont1 = org1.get_content()
        cont2 = org2.get_content()
        size1 = len(cont1)
        size2 = len(cont2)
        minsz = min(size1, size2)
        cut_point = random.randint(0, minsz-1)
        new_cont = cont1[:cut_point] + cont2[cut_point:]
        new_org = Organism(self)
        new_org.content = new_cont
        return new_org

    def generate_initial_population(self):
        for x in xrange(self.popCount):
            new_org = self.mutate(self.initial)
            self.commit(new_org)
            self.pop.append(new_org)

        a = self.pop.organisms[0]
        b = self.pop.organisms[1]

        self.pop.sort()
        for k in xrange(self.popCount):
            org = self.pop.organisms[k]
            print "%s ---  %s" % (org.content, org.fitness)

    def next_generation(self):
        new_pop = Population(self)
        for x in xrange(self.popCount):
            org1 = self.pop.select()
            org2 = self.pop.select()
            child = self.cross_over(org1, org2)
            if( random.random() < self.mutation_probability ):
                child = self.mutate(child)
            self.commit(child)
            new_pop.append(child)
        self.pop = new_pop
        self.pop.sort()
        for k in xrange(self.popCount):
            org = self.pop.organisms[k]
            print "%s ---  %s" % (org.content, org.fitness)
        

            
####################################################################################################################
    
if __name__ == "__main__":
    print
    print "-= empieza =-"
    example = 1

    if example == 1:
        directory = 'c:\\home\\fuzzing\\gph1\\'
        execfile = 'c:\\home\\fuzzing\\gph1\\fuzzme.exe'
        datafile = 'data.txt'
    if example == 2:
        directory = 'c:\\home\\fuzzing\\gph2\\' 
        execfile = 'c:\\home\\fuzzing\\gph2\\b2.exe'
        datafile = 'data.txt'

    print directory
    
    gm = G_Phazz_modifier()
    gm.load_executable(execfile)
    gm.load_initial_content(directory, datafile)
    
    gm.generate_initial_population()   

    for x in xrange(20):
        print "=========== generacion %s ============" % x
        gm.next_generation()
        print "fitness total = %s" % gm.pop.total_fitness
               
    print "-= chau! =-"

